;;; future.el --- general futures facility

;; Copyright (C) 2011  Teodor Zlatanov

;; Author: Teodor Zlatanov <tzz@lifelogs.com>
;; Keywords: data

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;; Code:

(eval-when-compile (require 'cl))

(defstruct future callback errorback status value)

(defun future-done-p (future)
  (future-status future))

(defun future-completed-p (future)
  (eq (future-status future) t))

(defun future-errored-p (future)
  (eq (future-status future) 'error))

(defun future-cancelled-p (future)
  (eq (future-status future) 'cancel))

(defun future-finish (future &optional status)
  (if (future-done-p future)
      (signal 'future-already-done nil)
    (setf (future-status future) (or status t))
    (when (future-callback future)
      (funcall (future-callback future) future))))

(defun future-errored (future error &optional signal)
  (if (future-done-p future)
      (signal 'future-already-done nil)
    (setf (future-status future) 'error)
    (setf (future-value future) signal)
    (when (future-errorback future)
      (funcall (future-errorback future) future error))))

(defun future-call (future)
  (if (future-done-p future)
    (let ((ff (future-value future)))
      (when (functionp ff)
        ;; TODO: needs error handling and to call future-errored
        (setf (future-value future)
              (funcall ff))
        (future-value future)))
    (future-finish future)))

(defun future-cancel (future)
  (if (future-done-p future)
      (signal 'future-already-done nil)
    (future-finish future 'cancel)))

(provide 'future)
;;; future.el ends here
